#include <stdio.h>
#include <jni.h>
#include <picl.h>
#include <malloc.h>
#include <string.h>
#include <alloca.h>
#include <inttypes.h>

#include "uk_co_petertribble_picl_PiclObject.h"
/* Header for class uk_co_petertribble_picl_JPiclObject */

/*
 * definitions imported from prtpicl.c
 */
#define	PRIxPICLTBL		PRIx64
#define	PRIxPICLHDL		PRIx64
#define	EM_USAGE		0
#define	EM_INIT			1
#define	EM_GETROOT		2
#define	EM_GETPVAL		3
#define	EM_GETNXTBYCOL		4
#define	EM_GETNXTBYROW		5
#define	EM_GETPINFO		6
#define	EM_GETPVALBYNAME	7
#define	EM_GETPROPBYNAME	8
#define	EM_INT_INVSIZE		9
#define	EM_UINT_INVSIZE		10
#define	EM_FLOAT_INVSIZE	11
#define	EM_TS_INVALID		12
#define	EM_TABLE_INVSIZE	13
#define	EM_REF_INVSIZE		14
#define	EM_TYPE_UNKNOWN		15
#define	EM_TS_OVERFLOW		16

#define JBUFSIZE 1024

static jclass po_class;
static jmethodID po_addnames_mid;
static jmethodID po_adddata_mid;
static jmethodID po_addblankdata_mid;
static jmethodID po_addchild_mid;
static jmethodID po_constructor_mid;

/*
 * print a bytearray value and format it to fit in 80 columns
 */
static void
jprint_bytearray(char *buf, uint8_t *vbuf, size_t nbytes)
{
	int		cnum;
	size_t		i;
	char		s[5];

	buf[0] = '\0';

	/*
	 * The COLUMNS_PER_BYTE is set to 4 to match the printf
	 * format used below, i.e. " %02x ", to print a byte
	 */
#define	COLUMNS_PER_BYTE	4
#define COLUMNS_FOR_OUTPUT	80

	cnum = 0;
	for (i = 0; i < nbytes; ++i) {
		if (cnum > COLUMNS_FOR_OUTPUT - COLUMNS_PER_BYTE) {
			(void) strlcat(buf, "\n", JBUFSIZE);
			cnum = 0;
		}
		(void) sprintf(s, " %02x ", vbuf[i]);
		(void) strlcat(buf, s, JBUFSIZE);
		cnum += COLUMNS_PER_BYTE;
	}
}


/*
 * Write a property's value into a String
 * If the property is read protected, return success.
 * If an invalid/stale handle error is encountered, return the error. For
 * other errors, print a message and return success.
 */
static int
jpicl_print_propval(picl_prophdl_t proph, const picl_propinfo_t *propinfo, char *buf)
{
	int		err;
	void		*vbuf;
	char		*str;
	uint64_t	val64;
	time_t		tmp;

	/*
	 * If property is read protected, print a message and continue
	 */
	if (!(propinfo->accessmode & PICL_READ)) {
		(void) sprintf(buf, "<%s>", "WRITE-ONLY");
		return (PICL_SUCCESS);
	}

	vbuf = alloca(propinfo->size);
	if (propinfo->type == PICL_PTYPE_VOID)
		return (PICL_SUCCESS);

	err = picl_get_propval(proph, vbuf, propinfo->size);
	/*
	 * If the error is not a stale/invalid handle or noresponse, continue
	 * by ignoring the error/skipping the property.
	 */
	if ((err == PICL_INVALIDHANDLE) || (err == PICL_STALEHANDLE) ||
	    (err == PICL_NORESPONSE))
		return (err);
	else if (err != PICL_SUCCESS) {
	        (void) sprintf(buf, "<%s>", "ERROR");
		return (PICL_SUCCESS);
	}

	switch (propinfo->type) {
	case PICL_PTYPE_CHARSTRING:
		(void) snprintf(buf, JBUFSIZE, "%s", (char *)vbuf);
		break;
	case PICL_PTYPE_INT:
		switch (propinfo->size) {
		case sizeof (int8_t):
			/* avoid using PRId8 until lint recognizes hh */
			(void) sprintf(buf, " %d ", *(int8_t *)vbuf);
			break;
		case sizeof (int16_t):
			(void) sprintf(buf, " %" PRId16 " ", *(int16_t *)vbuf);
			break;
		case sizeof (int32_t):
			(void) sprintf(buf, " %" PRId32 " ", *(int32_t *)vbuf);
			break;
		case sizeof (int64_t):
			(void) sprintf(buf, " %" PRId64 " ", *(int64_t *)vbuf);
			break;
		default:
		  /*
			print_errmsg(gettext(err_msg[EM_INT_INVSIZE]),
			    propinfo->size);
			    */
			return (PICL_FAILURE);
		}
		break;
	case PICL_PTYPE_UNSIGNED_INT:
		switch (propinfo->size) {
		case sizeof (uint8_t):
			/* avoid using PRIx8 until lint recognizes hh */
			(void) sprintf(buf, " %#x ", *(uint8_t *)vbuf);
			break;
		case sizeof (uint16_t):
			(void) sprintf(buf, " %#" PRIx16 " ", *(uint16_t *)vbuf);
			break;
		case sizeof (uint32_t):
			(void) sprintf(buf, " %#" PRIx32 " ", *(uint32_t *)vbuf);
			break;
		case sizeof (uint64_t):
			(void) sprintf(buf, " %#" PRIx64 " ", *(uint64_t *)vbuf);
			break;
		default:
			return (PICL_FAILURE);
		}
		break;
	case PICL_PTYPE_FLOAT:
		(void) sprintf(buf, "%s", "A floating number");
		switch (propinfo->size) {
		case sizeof (float):
			(void) printf(" %f ", *(float *)vbuf);
			break;
		case sizeof (double):
			(void) printf(" %f ", *(double *)vbuf);
			break;
		default:
			return (PICL_FAILURE);
		}
		break;
	case PICL_PTYPE_TIMESTAMP:
		val64 = *(uint64_t *)vbuf;
		tmp = (time_t)val64;
		if ((uint64_t)tmp != val64) {
			return (PICL_FAILURE);
		}
		str = ctime(&tmp);
		if (str == NULL) {
			return (PICL_FAILURE);
		}
		str[strlen(str) - 1] = '\0';
		(void) sprintf(buf, " %s ", str);
		break;
	case PICL_PTYPE_TABLE:
		(void) sprintf(buf, "%s", "A table");
		/*
		(void) printf("%s\n", "A table");
		*/
		if (propinfo->size != sizeof (picl_prophdl_t)) {
			return (PICL_FAILURE);
		}
		/*
		(void) printf("(%" PRIxPICLTBL "TBL) ",
		    *(picl_prophdl_t *)vbuf);
		*/
		break;
	case PICL_PTYPE_REFERENCE:
	        /*
		(void) printf("%s", "A reference");
		*/
		(void) sprintf(buf, "%s", "A reference");
		if (propinfo->size != sizeof (picl_nodehdl_t)) {
			return (PICL_FAILURE);
		}
		/*
		(void) printf(" (%" PRIxPICLHDL "H) ", *(picl_nodehdl_t *)vbuf);
		*/
		break;
	case PICL_PTYPE_BYTEARRAY:
		jprint_bytearray(buf, vbuf, propinfo->size);
		break;
	default:
		(void) sprintf(buf, "<%s>", "UNKNOWN TYPE");
		return (PICL_FAILURE);
	}
	return (PICL_SUCCESS);
}

/*
 * add the picl data for this node
 * this is the print_proplist routine from prtpicl.c
 */
int jpicl_add_proplist(JNIEnv *env, picl_nodehdl_t nodeh, jobject jobj)
{
	int		err;
	picl_prophdl_t	proph;
	picl_propinfo_t	propinfo;
	picl_prophdl_t	tblh;
	jstring jpname;
	jstring jpvalue;
	char buf[JBUFSIZE];
	char nbuf[JBUFSIZE];

	for (int nerr = picl_get_first_prop(nodeh, &proph); nerr == PICL_SUCCESS;
	    nerr = picl_get_next_prop(proph, &proph)) {

		err = picl_get_propinfo(proph, &propinfo);
		if (err != PICL_SUCCESS) {
			return (err);
		}

		(void) snprintf(nbuf, JBUFSIZE, "%s", propinfo.name);
		if (propinfo.type == PICL_PTYPE_VOID) {
			jpname = (*env)->NewStringUTF(env, nbuf);
			(*env)->CallVoidMethod(env, jobj, po_addblankdata_mid, jpname);
			(*env)->ReleaseStringUTFChars(env, jpname, nbuf);
		}
		else {
			jpname = (*env)->NewStringUTF(env, nbuf);
			err = jpicl_print_propval(proph, &propinfo, buf);
			if (err != PICL_SUCCESS) {
			  (*env)->CallVoidMethod(env, jobj, po_addblankdata_mid
, jpname);
			} else {
			  jpvalue = (*env)->NewStringUTF(env, buf);
		/*
		(void) printf("%s\n", "in add_proplist call, copied");
		(void) printf("value %s\n", buf);
		(void) printf("name %s\n", nbuf);
		*/
			  (*env)->CallVoidMethod(env, jobj, po_adddata_mid, jpname, jpvalue);
		/*
		(void) printf("%s\n", "in add_proplist call, called method");
		*/
			  (*env)->ReleaseStringUTFChars(env, jpvalue, buf);
			}
			(*env)->ReleaseStringUTFChars(env, jpname, nbuf);
			if (err != PICL_SUCCESS)
				return (err);
		}

		/*
		 * Expand the table property
		 */
		if (propinfo.type == PICL_PTYPE_TABLE) {
			err = picl_get_propval(proph, &tblh, propinfo.size);
			if (err != PICL_SUCCESS) {
				return (err);
			}
			/*
			 * FIXME implement
			err = print_table_prop(lvl, tblh);
			*/
			if (err != PICL_SUCCESS)
				return (err);
		}
	}
	return (PICL_SUCCESS);
}

int jpicl_walk_tree(JNIEnv *env, picl_nodehdl_t nodeh, jobject jobj)
{
	int err;
	picl_nodehdl_t	chdh;
	picl_prophdl_t	proph;
	picl_propinfo_t	pinfo;
	char		classval[PICL_PROPNAMELEN_MAX];
	char		cnameval[PICL_PROPNAMELEN_MAX];
	jstring jcstr;
	jstring jnstr;
	jobject njobj;
	err = picl_get_propval_by_name(nodeh, PICL_PROP_CLASSNAME, classval,
	    sizeof (classval));
	err = picl_get_propval_by_name(nodeh, PICL_PROP_NAME, cnameval,
	    sizeof (cnameval));
	jcstr = (*env)->NewStringUTF(env, classval);
	jnstr = (*env)->NewStringUTF(env, cnameval);
	(*env)->CallVoidMethod(env, jobj, po_addnames_mid, jcstr, jnstr);
	/* 
	 * fill in the data points for this node
	 * corresponds to calling print_proplist in prtpicl
	 */
	/*
	(void) printf("%s\n", "Calling add_proplist");
	*/
	err = jpicl_add_proplist(env, nodeh, jobj);
	/*
	(void) printf("%s\n", "Returned from add_proplist");
	*/
	if (err != PICL_SUCCESS)
		return (err);
	/* look for and iterate over children */
	for (err = picl_get_propval_by_name(nodeh, PICL_PROP_CHILD, &chdh,
	    sizeof (picl_nodehdl_t)); err != PICL_PROPNOTFOUND;
		err = picl_get_propval_by_name(chdh, PICL_PROP_PEER, &chdh,
		    sizeof (picl_nodehdl_t))) {

		if (err != PICL_SUCCESS)
			return (err);

		njobj = (*env)->NewObject(env, po_class, po_constructor_mid);
		(*env)->CallVoidMethod(env, jobj, po_addchild_mid, njobj);
		err = jpicl_walk_tree(env, chdh, njobj);
		if (err != PICL_SUCCESS)
			return (err);
	}

	(*env)->ReleaseStringUTFChars(env, jcstr, classval);
	(*env)->ReleaseStringUTFChars(env, jnstr, cnameval);

	return(PICL_SUCCESS);
}

JNIEXPORT jint JNICALL Java_uk_co_petertribble_picl_PiclObject_enumerate
  (JNIEnv *env, jobject jobj)
{
  int err;
  picl_nodehdl_t	rooth;
	err = picl_initialize();
	if (err != PICL_SUCCESS) {
	  return(NULL);
	}
	err = picl_get_root(&rooth);
	if (err != PICL_SUCCESS) {
	  return(NULL);
	}
	po_class = (*env)->FindClass(env, "uk/co/petertribble/picl/PiclObject");
	po_constructor_mid = (*env)->GetMethodID(env, po_class, "<init>",
            "()V");
	po_addnames_mid = (*env)->GetMethodID(env, po_class, "setNames",
            "(Ljava/lang/String;Ljava/lang/String;)V");
	po_adddata_mid = (*env)->GetMethodID(env, po_class, "addData",
            "(Ljava/lang/String;Ljava/lang/String;)V");
	po_addblankdata_mid = (*env)->GetMethodID(env, po_class, "addBlankData",
            "(Ljava/lang/String;)V");
	po_addchild_mid = (*env)->GetMethodID(env, po_class, "addChild",
            "(Luk/co/petertribble/picl/PiclObject;)V");
	err = jpicl_walk_tree(env, rooth, jobj);
	(void) picl_shutdown();
	return(err);
}
